module samples/telesystem
open std/ord

sig Person {}

sig State {
	delegate: Person->?Person,
	followme: Person->?Person,
	// these three are variables used to visualize 
	// the results of the translation functions in the Graph
	final_del: Person->Person,
	final_fol: Person->Person,
	final: Person->Person
	}

fact SetupInfo {
	all s: State, p: Person {
		s.final_del[p] = DelegateDest( s, p )
		s.final_fol[p] = FollowmeDest( s, p )
		s.final    [p] = Dest        ( s, p )
		}
	}

fact SomePeople {
	Person = univ[Person]
	State = univ[State]
	}

fun DelegateDest( s: State, p: Person ): Person {
	result = *(s.delegate)[p] - (s.delegate).( *(s.delegate)[p] )
	}

fun FollowmeDest( s: State, p: Person ): Person {
	result = if some s.followme[p] then s.followme[p] else p
	}

fun Dest( s: State, p: Person ): Person {
	result = FollowmeDest( s, DelegateDest( s, p ) )
	}

//New Constraints Needed
//fixes NotAllDelegate
fun NoCircularDelegateReference() {
	all s: State, p: Person | p not in ^(s.delegate)[p]
	}
//fixes AllLinksUsed
fun NoDelegateAndFollowme() {
	all s: State, p: Person {
		not (some s.delegate[p] and some s.followme[p])
		}
	}


//Checks and Assertions
fun OneDestination() {
	all s: State, p: Person | one Dest( s, p )
	}

assert OneDestination_NoLoops {
	NoCircularDelegateReference() => OneDestination()
	}
check OneDestination_NoLoops for 3 but 1 State

// this solves the OneDestination problem as well
assert OneDestination_TotalOrdering {
	( all p: Person | State.delegate[p] in OrdPrevs(p)-p )
	=> OneDestination()
	}
check OneDestination_TotalOrdering for 3 but 1 State

assert NotAllDelegate {   
	NoCircularDelegateReference() =>
	( some s: State, p: Person | p = DelegateDest( s, p ) )
	}
check NotAllDelegate for 3 but 1 State


fun GetUsedFollowmeMappings( s: State ): Person->Person {
	all p: DelegateDest( s, Person ) {
		p->FollowmeDest( s, p ) in result
		}
	}

assert AllLinksUsed {
	( NoCircularDelegateReference() && NoDelegateAndFollowme() ) =>
	=>
	( all s: State {
		all fm: s.followme {
			fm in GetUsedFollowmeMappings( s )
			}
		} )
	}
check AllLinksUsed for 3 but 1 State

fun BaseChecks() {
	NoDelegateAndFollowme()
	NoCircularDelegateReference()
	}
//Running Examples

fun AllFollow() {
	BaseChecks()
	all s: State, p: Person | p != FollowmeDest( s, p )
	}
run AllFollow for 4 but 1 State

fun AllDelegateButOne() {
	BaseChecks()
	all s: State {
		one p: Person | p = DelegateDest( s, p )
		}
	}
run AllDelegateButOne for 3 but 1 State


fun ChangePerTransition() {
	BaseChecks()
	all s: State - Ord[State].last {
		one p: Person {
			OrdNext(s).delegate[p] != s.delegate[p]
			or
			OrdNext(s).followme[p] != s.followme[p]
			}
		} 
	}
run ChangePerTransition for 3 but 2 State

/*
part sig Alice, Bob, Carl, David extends Person {}
fun ABCDExample() {
	BaseChecks()
	#Alice = 1
	#Bob = 1
	#Carl = 1
	#David = 1
	State.followme[Bob] = Carl
	State.followme[Carl] = David
	State.delegate[Alice] = Bob
	}
run ABCDExample for 4 but 1 State
*/