module samples/telesystemtime
open std/ord


sig Person {}
sig AnsweringMachine extends Person {}

sig Status {}
part sig Available, Unavailable extends Status {}
fact StatusFact {
	#Available = 1
	#Unavailable = 1
	#Status = 2
	}

sig Call {}

sig State {
	entitystatus: Person->!Status,
	entityforward: Person->!Person,
	callstatus: Call?->?Person,
	callscompleted: Call->?Person,
	callsfailed: Call->?Person
	}

fact CallLogistics {
	all c: Call | some State.callstatus[c]
	no Ord[State].first.callscompleted[Call]
	no Ord[State].first.callsfailed[Call]
	}

fact AnsweringMachineConstraints {
	State.entitystatus[AnsweringMachine] in Available
	no State.entityforward[AnsweringMachine]
	}

fact useAll {
	univ[Person] = Person
	univ[Call] = Call
	univ[Status] = Status
	}

fact AllPeopleHaveStatus {
	all p: Person, s: State | one s.entitystatus[p]
	}

fact CallOnlyInOnePlace {
	all s: State, c: Call {
		#s.callstatus[c] + #s.callscompleted[c] + #s.callsfailed[c] < 2
		}
	}


fun Transition( s, s': State ) {
	//Progress all calls, or end them
	all c: (s.callstatus).Person {
		s.entitystatus[ s.callstatus[c] ] = Unavailable => {
			some s.entityforward[ s.callstatus[c] ] =>
				s'.callstatus[c] = s.entityforward[ s.callstatus[c] ]
			no s.entityforward[ s.callstatus[c] ] => {
				no s'.callstatus[c]
				s'.callsfailed[c] = s.entityforward[ s.callstatus[c] ]
				}
			}
		s.entitystatus[ s.callstatus[c] ] = Available => {
			no s'.callstatus[c]
			s'.callscompleted[c] = s.entityforward[ s.callstatus[c] ]
			s'.entitystatus[ s.callstatus[c] ] = Unavailable
			}
		}
	s.callsfailed[Call] in s'.callsfailed[Call]
	s.callscompleted[Call] in s'.callscompleted[Call]
	}

fact Constraints {
	all s: State - Ord[State].last {
		Transition( s, OrdNext(s) )
		}
	}

// No complete failure loops in one state
fun NoLoopsInState() {
	all s: State | all p: Person {
		p not in ^(s.entityforward)[p] or 
		some p': ^(s.entityforward)[p] | s.entitystatus[p'] = Available
		}
	}

// At most one new call
fun OneNewCallPerTick( s, s': State ) {
	//Add at most one new call each transition
	one p: Person | s'.callstatus[Call] - p in s.callstatus[Call]
	}

// One person must change their status each time
fun OneStatusChangePerTick( s, s': State ) {
	one p: Person | s'.entitystatus[p] != s.entitystatus[p]
	}

// Only one person can change their status each time
fun MaybeOneStatusChangePerTick( s, s': State ) {
	sole p: Person | s'.entitystatus[p] != s.entitystatus[p]
	}

// Do not allow changes to the forwarding table
fun NoForwardingChangesPerTick( s, s': State ) {
	all p: Person | s'.entityforward[p] = s.entityforward[p]
	}


fun LoopWithoutProblems_FullLengthCallTraversal() {
	NoLoopsInState()
	(Ord[State].first.callstatus).Person = Call
	some OrdPrev( Ord[State].last ).callstatus[Call]
	(Ord[State].last.callscompleted).Person = Call
	all s: State {
		some p: Person | ^(s.entityforward)[p] = Person
		}
	all s: State - Ord[State].last {
		NoForwardingChangesPerTick( s, OrdNext(s) )
		MaybeOneStatusChangePerTick( s, OrdNext(s) )
		}
	}

run LoopWithoutProblems_FullLengthCallTraversal for 3 but 3 State, 1 Call, 2 Status

//Demonstrates that with 1 person, one can have infinite loops if direction is allowed to oneself
fun RunLongTime() {
	some Ord[State].first.callstatus[Call]
	some Ord[State].last.callstatus[Call]
	all s: State - Ord[State].last {
		NoForwardingChangesPerTick( s, OrdNext(s) )
		MaybeOneStatusChangePerTick( s, OrdNext(s) )
		}
	}

run RunLongTime for 1 but 6 State, 1 Call, 2 Status

//Demonstrates that with 1 person, one doesn't need infinite loops if direction is allowed to oneself
fun RunShortTime() {
	some Ord[State].first.callstatus[Call]
	some Ord[State].last.callscompleted[Call]
	all s: State - Ord[State].last {
		NoForwardingChangesPerTick( s, OrdNext(s) )
		MaybeOneStatusChangePerTick( s, OrdNext(s) )
		}
	}

run RunShortTime for 1 but 6 State, 1 Call, 2 Status

//Demonstrates that with 4 person, and no loops at any point in time, one can still have infinite loops for the Call
fun RunLongTimeWithNoLoops() {
	NoLoopsInState()
	some Ord[State].first.callstatus[Call]
	some Ord[State].last.callstatus[Call]
	all s: State - Ord[State].last {
		NoForwardingChangesPerTick( s, OrdNext(s) )
		}
	}

run RunLongTimeWithNoLoops for 2 but 4 State, 1 Call, 2 Status

assert RunLongTimeWithNoLoops_AndOneStatusChangePerTick {
	RunLongTimeWithNoLoops()
	all s: State - Ord[State].last {
		MaybeOneStatusChangePerTick( s, OrdNext(s) )
		}
	some s: State - Ord[State].last | Ord[State].last = s
	}

check RunLongTimeWithNoLoops_AndOneStatusChangePerTick for 3 but 12 State, 1 Call, 2 Status
